from ._base import SarvModule
from ._mixins import UrlMixin

class Products(SarvModule, UrlMixin):
    _module_name = 'AOS_Products'
    _table_name = 'aos_products'
    _assigned_field = 'assigned_user_id'
    _label_en = 'Products'
    _label_pr = 'محصولات'


from classmods import logwrap
from typing import Any, Optional
from ._base import SarvModule
from ._mixins import UrlMixin

class Users(SarvModule, UrlMixin):
    _module_name = 'Users'
    _table_name = 'users'
    _assigned_field = 'assigned_user_id'
    _label_en = 'Users'
    _label_pr = 'کاربران'

    @logwrap(before='Reading currnet user detail: args:{args} - kwargs:{kwargs}', after=False)
    def get_me(
            self, 
            selected_fields: Optional[list[str]] = None,
            caching: bool = False,
            expire_after: int = 300,
        ) -> dict[str, Any]:
        """
        Returns current user details.

        Args:
            selected_fields (list[str], optional): A list of fields to include in the response.
            caching (bool, optional): Whether to cache the results.
            expire_after (int, optional): The time in seconds to cache the results.
        
        Returns:
            dict: Users data
        """
        return self.read_list(
            query=f"users.user_name='{self._client._username.lower()}'",
            selected_fields=selected_fields,
            limit=1,
            caching=caching,
            expire_after=expire_after,
        )[0]


from ._base import SarvModule
from .accounts import Accounts
from .acl_roles import ACLRoles
from .aos_contracts import SalesContracts
from .aos_invoices import Invoices
from .aos_pdf_templates import PDFTemplates
from .aos_product_categories import ProductCategories      
from .aos_products import Products
from .aos_quotes import Quotes
from .approval import Approvals
from .asol_project import Projects
from .branches import Branches
from .bugs import Bugs
from .calls import Calls
from .cases import Cases
from .communications import Communications
from .communications_target import CommunicationTargets      
from .communications_template import CommunicationTemplates   
from .campaigns import Campaigns
from .contacts import Contacts
from .deposits import Deposits
from .documents import Documents
from .emails import Emails
from .knowledge_base import KnowledgeBases
from .knowledge_base_categories import KnowledgeBaseCategories
from .leads import Leads
from .meetings import Meetings
from .notes import Notes
from .notifications import Notifications
from .obj_conditions import Conditions
from .obj_indicators import Indicators
from .obj_objectives import Objectives
from .opportunities import Opportunities
from .payments import Payments
from .purchase_order import PurchaseOrders
from .sc_competitor import Competitor
from .sc_contract import SupportContracts
from .sc_contract_management import Services
from .service_centers import ServiceCenters
from .tasks import Tasks
from .timesheet import Timesheets
from .users import Users
from .vendors import Vendors


__all__ = [
    'SarvModule',
    'Accounts',
    'ACLRoles',
    'SalesContracts',
    'Invoices',
    'PDFTemplates',
    'ProductCategories',
    'Products',
    'Quotes',
    'Approvals',
    'Projects',
    'Branches',
    'Bugs',
    'Calls',
    'Cases',
    'Communications',
    'CommunicationTargets',
    'CommunicationTemplates',
    'Campaigns',
    'Contacts',
    'Deposits',
    'Documents',
    'Emails',
    'KnowledgeBases',
    'KnowledgeBaseCategories',
    'Leads',
    'Meetings',
    'Notes',
    'Notifications',
    'Conditions',
    'Indicators',
    'Objectives',
    'Opportunities',
    'Payments',
    'PurchaseOrders',
    'Competitor',
    'SupportContracts',
    'Services',
    'ServiceCenters',
    'Tasks',
    'Timesheets',
    'Users',
    'Vendors',
]


from ._base import SarvModule
from ._mixins import UrlMixin

class ACLRoles(SarvModule, UrlMixin):
    _module_name = 'ACLRoles'
    _table_name = 'acl_roles'
    _assigned_field = ''
    _label_en = 'ACL Roles'
    _label_pr = 'نقش ها'


import json, hashlib, requests, requests_cache, urllib.parse
from classmods import ENVMod, logwrap
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Literal, Optional, Self
from .modules import SarvModule
from ._exceptions import SarvException
from ._mixins import ModulesMixin
from ._type_hints import TimeOutputMethods, SarvLanguageType, RequestMethod, SarvGetMethods
from ._url import SarvURL, SarvFrontend


class SarvClient(ModulesMixin):
    """
    SarvClient provides methods for interacting with the SarvCRM API. 
    It supports authentication, data retrieval, and other API functionalities.
    """
    @ENVMod.register(section_name='SarvClient')
    @logwrap(before=('DEBUG','Initiating SarvClient'), after='Initiating compleated')
    def __init__(
            self,
            utype: str,
            username: str,
            password: str,
            is_password_md5: bool = False,
            url: Optional[str] = None,
            frontend_url: Optional[str] = None,
            login_type: Optional[str] = None,
            auto_login: bool = True,
            language: SarvLanguageType = 'en_US',
            caching: bool = False,
            cache_backend: Literal['memory', 'sqlite'] = 'memory',
        ) -> None:
        """
        Initialize the SarvClient.

        Args:
            utype (str): The user type for authentication.
            username (str): The username for authentication.
            password (str): The password for authentication.
            is_password_md5 (bool): Whether the password is already hashed using MD5.
            url (Optional[str]): The base URL for the SarvCRM API if you use local instance.
            frontend_url (Optional[str]): The base URL for the SarvCRM frontend if you use local instance.
            login_type (Optional[str]): The login type for authentication.
            auto_login (bool): Try to login once on status code 401, 403
            language (SarvLanguageType): The language to use, default is 'en_US'.
            caching (bool): Cache the responses from server.
            cache_backend(Literal['memory', 'sqlite']): Saving backend for cached responses.
        """
        self._utype = utype
        self._username = username
        self._password = password if is_password_md5 else self.hash_password(password)
        self._url = url or SarvURL
        self._frontend_url = frontend_url or SarvFrontend
        self._login_type = login_type
        self._auto_login = auto_login
        self._language = language
        self._caching = caching
        self._cache_backend = cache_backend

        self._token: str = ''

        self.enable_caching() if self._caching else self.disable_caching()

        super().__init__()
        self._user_id: str | None = None

    @property
    def user_id(self) -> str:
        if not self._user_id:
            self._user_id = self.Users.get_me(selected_fields=['id']).get('id', '')

        assert self._user_id, 'Unable to retrive user_id from server.'
        return self._user_id

    @logwrap(before=False, after=('DEBUG','Added Headers to Session'))
    def _add_headers(self) -> None:
        """
        Adds required sarvcrm headers to session.
        """
        self._session.headers.update({'Content-Type': 'application/json'})
        self._session.headers.update({'Accept': 'application/json'})

    @logwrap(before=False, after=('INFO', 'Enabled caching'))
    def enable_caching(self) -> None:
        """
        Enables the caching and replaces `Session` with `CachedSession` or creates it.
        """
        self._session = requests_cache.CachedSession(
            cache_name='sarv_api_cache',
            backend=self._cache_backend,
            allowable_methods=('GET', 'POST'),
            allowable_codes=(200,),
        )
        self._add_headers()

    @logwrap(before=False, after=('INFO', 'Disabled caching'))
    def disable_caching(self) -> None:
        """
        Disables the caching and replaces `CachedSession` with `Session` or creates it.
        """
        self._session = requests.Session()
        self._add_headers()

    @logwrap(before='Sending HTTP request: args:{args} - kwargs:{kwargs}', after=False)
    def _send_request(
            self, 
            request_method: RequestMethod,
            endpoint: Optional[str] = None,
            head_params: Optional[dict] = None,
            get_params: Optional[dict] = None,
            post_params: Optional[dict] = None,
            auto_login: Optional[bool] = None,
            caching: bool = False,
            expire_after: int = 300,
        ) -> Any:
        """
        Send a request to the Sarv API and return the response data.

        Args:
            request_method (RequestMethod): The HTTP method for the request ('GET', 'POST', etc.).
            head_parms (dict): The headers for the request.
            get_parms (dict): The GET parameters for the request.
            post_params (dict): The POST parameters for the request.
            auto_login (Optional(bool)): Whether to auto login on status code 401, 403.
            caching (bool, optional): Whether to cache the results.
            expire_after (int, optional): The time in seconds to cache the results.

        Returns:
            Any: The data parameter from the server response that can be `List` or `Dict`

        Raises:
            SarvException: If the server returns an error response.
        """
        head_params = head_params or {}
        get_params = get_params or {}
        post_params = post_params or {}

        head_params = {k: v for k, v in head_params.items() if v is not None} 
        get_params = {k: v for k, v in get_params.items() if v is not None}
        post_params = {k: v for k, v in post_params.items() if v is not None}

        auto_login = auto_login if auto_login is not None else self._auto_login

        if self._token:
            head_params['Authorization'] = f'Bearer {self._token}'

        kwargs = {
            'method': request_method,
            'url': urllib.parse.urljoin(self._url.rstrip('/') + '/', (endpoint or '').lstrip('/')),
            'headers': head_params,
            'params': get_params,
            'json': post_params,
            'verify': True,
        }

        if isinstance(self._session, requests_cache.CachedSession):
            ## Use cache
            if caching:
                kwargs.update({'expire_after': expire_after})
                response: requests.Response = self._session.request(**kwargs)

            ## If caching enabled but user dont want to use it
            else:
                with self._session.cache_disabled():
                    response: requests.Response = self._session.request(**kwargs)

        else:
            ## Normal request without caching
            response: requests.Response = self._session.request(**kwargs)

        # Check for Server respond
        try:
            # Deserialize sarvcrm servers response
            response_dict: dict = response.json()

        # This is for quirky responses from Sarvcrm servers
        # Sometimes servers send other content types instead of json
        except json.decoder.JSONDecodeError:
            if 'MySQL Error' in response.text:
                raise SarvException(
                    'There are Errors in the database\n'
                    'if you are sending raw SQL Query to server\n'
                    'please check syntax and varible names'
                )
            else:
                raise SarvException(
                    'Unkhown error from Server while parsing json'
                )

        try:
            response.raise_for_status()

        except requests.HTTPError as e:
            if not (auto_login and response.status_code in (401, 403)):
                raise e

            head_params.pop('Authorization', None)
            self.login()

            return self._send_request(
                request_method=request_method,
                endpoint=endpoint,
                head_params=head_params,
                get_params=get_params,
                post_params=post_params,
                auto_login=False,
                caching=False,
            )

        return response_dict.get('data', {})

    @logwrap(before=('INFO', 'Logging to Sarvcrm'), after=False)
    def login(self) -> str:
        """
        Authenticate the user and retrieve an access token.

        Returns:
            str: The access token for authenticated requests.
        """
        post_params = {
            'utype': self._utype,
            'user_name': self._username,
            'password': self._password,
            'login_type': self._login_type,
            'language': self._language,
        }

        response: Dict[str, Any] = self._send_request(
            request_method='POST',
            get_params=self._create_get_params('Login'), 
            post_params=post_params,
            caching=False,
            auto_login=False,
        )

        self._token = response['token']

        return self._token

    @logwrap(before=('INFO', 'Logging out from Sarvcrm'), after=False)
    def logout(self) -> None:
        """
        Clears the access token from the instance.

        This method should be called to invalidate the session.
        """
        if self._token:
            self._token = ''

    @logwrap(before='Searching by number: args:{args} - kwargs:{kwargs}', after=False)
    def search_by_number(
            self,
            number: str,
            module: Optional[SarvModule | str] = None,
            caching: bool = False,
            expire_after: int = 300,
        ) -> List[Dict[str, Any]]:
        """
        Search the CRM by phone number and retrieve the module item.

        Args:
            number (str): The phone number to search for.
            module (Optional[SarvModule | str]): The module to search in.
            caching (bool, optional): Whether to cache the results.
            expire_after (int, optional): The time in seconds to cache the results.

        Returns:
            dict: The data related to the phone number if found.
        """
        return self._send_request(
            request_method = 'GET',
            get_params = self._create_get_params(
                'SearchByNumber', 
                sarv_module = module, 
                number = number,
                caching = caching,
                expire_after = expire_after,
            ),
        )

    @staticmethod
    def _create_get_params(
            sarv_get_method: Optional[SarvGetMethods] = None,
            sarv_module: Optional[SarvModule | str] = None,
            **addition
        ) -> Dict[str, Any]:
        """
        Create the GET parameters with the method and module.

        Args:
            sarv_get_method (SarvGetMethods): The API method to call.
            sarv_module (Optional[SarvModule | str]): The module name or object.
            addition: Additional parameters to include in the GET request.

        Returns:
            dict: The constructed GET parameters.
        """
        module_name = None

        if sarv_module is not None:
            if isinstance(sarv_module, SarvModule):
                module_name = sarv_module._module_name
            elif isinstance(sarv_module, str):
                module_name = sarv_module
            else:
                raise TypeError(f'Module type must be instance of SarvModule or str not {sarv_module.__class__.__name__}')

        get_parms = {
            'method': sarv_get_method,
            'module': module_name,
        }

        if addition:
            get_parms.update(**addition)

        return get_parms

    @staticmethod
    def iso_time_output(
            output_method: TimeOutputMethods,
            dt: datetime | timedelta,
        ) -> str:
        """
        Generate a formatted string from a datetime or timedelta object.

        These formats are compliant with the SarvCRM API time standards.

        Args:
            output_method (TimeOutputMethods): Determines the output format ('date', 'datetime', or 'time').
            dt (datetime | timedelta): A datetime or timedelta object.

        Returns:
            str: A string representing the date, datetime, or time.
                - date: "YYYY-MM-DD"
                - datetime: "YYYY-MM-DDTHH:MM:SS+HH:MM"
                - time: "HH:MM:SS"
        """
        if isinstance(dt, timedelta):
            dt = datetime.now(timezone.utc) + dt

        if output_method == 'date':
            return dt.date().isoformat()

        elif output_method == 'datetime':
            return dt.astimezone().isoformat(timespec="seconds")

        elif output_method == 'time':
            return dt.time().isoformat(timespec="seconds")

        else:
            raise TypeError(f'Invalid output method: {output_method}')

    @staticmethod
    def hash_password(password: str) -> str:
        """
        Returns the acceptable hash for SarvCRM Login

        Args:
            password(str): your password
        
        Returns:
            str: md5 hashed password
        """
        return hashlib.md5(password.encode('utf-8')).hexdigest()

    @logwrap(before='Starting SarvClient context.', after=('DEBUG', 'SarvClient context started.'))
    def __enter__(self) -> Self:
        """Basic Context Manager for clean code execution"""
        if not self._token:
            self.login()

        return self

    @logwrap(before='Closing SarvClient context.', after=('DEBUG', 'SarvClient context closed.'))
    def __exit__(self, exc_type, exc_value, traceback):
        """Basic Context Manager for clean code execution"""
        self.logout()

    def __repr__(self):
        """
        Provides a string representation for debugging purposes.

        Returns:
            str: A string containing the class name and key attributes.
        """
        return f'{self.__class__.__name__}(utype={self._utype}, username={self._username})'

    def __str__(self) -> str:
        """
        Provides a human-readable string representation of the instance.

        Returns:
            str: A simplified string representation of the instance.
        """
        return f'<SarvClient {self._utype}-{self._username}>'


from .modules import *

class ModulesMixin:
    """This is abstract module for adding sarv modules to sarvclient class"""
    def __init__(self) -> None:
        self.Accounts = Accounts(self)
        self.ACLRoles = ACLRoles(self)
        self.SalesContracts = SalesContracts(self)
        self.Invoices = Invoices(self)
        self.PDFTemplates = PDFTemplates(self)
        self.ProductCategories = ProductCategories(self)
        self.Products = Products(self)
        self.Quotes = Quotes(self)
        self.Approvals = Approvals(self)
        self.Projects = Projects(self)
        self.Branches = Branches(self)
        self.Bugs = Bugs(self)
        self.Calls = Calls(self)
        self.Cases = Cases(self)
        self.Communications = Communications(self)
        self.CommunicationsTargets = CommunicationTargets(self)
        self.CommunicationsTemplates = CommunicationTemplates(self)
        self.Campaigns = Campaigns(self)
        self.Contacts = Contacts(self)
        self.Deposits = Deposits(self)
        self.Documents = Documents(self)
        self.Emails = Emails(self)
        self.KnowledgeBases = KnowledgeBases(self)
        self.KnowledgeBaseCategories = KnowledgeBaseCategories(self)
        self.Leads = Leads(self)
        self.Meetings = Meetings(self)
        self.Notes = Notes(self)
        self.Notifications = Notifications(self)
        self.Conditions = Conditions(self)
        self.Indicators = Indicators(self)
        self.Objectives = Objectives(self)
        self.Opportunities = Opportunities(self)
        self.Payments = Payments(self)
        self.PurchaseOrders = PurchaseOrders(self)
        self.Competitor = Competitor(self)
        self.SupportContracts = SupportContracts(self)
        self.Services = Services(self)
        self.ServiceCenters = ServiceCenters(self)
        self.Tasks = Tasks(self)
        self.Timesheets = Timesheets(self)
        self.Users = Users(self)
        self.Vendors = Vendors(self)
